/* for gcc */
#if defined _WIN32 || defined __CYGWIN__
#include <windows.h>
#endif

#include "ruby.h"
#include "regex.h"

/* Because ruby 1.6 doesn't have StringValuePtr macro, define here */
#ifndef StringValuePtr
#define StringValuePtr(v) rb_str2cstr(v, 0)
#endif

VALUE rb_mDOSish;

#define isdirsep(x) ((x) == '/' || (x) == '\\')
#ifndef CharNext		/* defined as CharNext[AW] on Windows. */
# if defined(DJGPP)
#   define CharNext(p) ((p) + mblen(p, MB_CUR_MAX))
# else
#   define CharNext(p) ((p) + mbclen(*p))
# endif
#endif

static char *strrdirsep _((const char *path));
static int extract_file_drive _((const char *d));
static int extract_file_root _((const char *d));
static void conv_separator _((char *p));

static char *
strrdirsep(path)
    const char *path;
{
    const char *last = NULL;
    while (*path) {
	if (isdirsep(*path)) {
	    last = path++;
	}
	else {
	    path = CharNext(path);
	}
    }
    return (char *)last;
}

static int
extract_file_drive(d)
    const char *d;
{
    const char *p;
    int len;
    
    if (d[0] && d[1] == ':') {
	return 2;
    }
    else if (isdirsep(d[0]) && isdirsep(d[1])) {
	int sepcnt = 0;
	const char *last = p = d + 2;
	while (*p) {
	    if (isdirsep(*p)) {
		last = p++;
		if (sepcnt++ == 1) break;
	    }
	    else {
		p = CharNext(p);
	    }
	}
	if (p-1 == last) p = last;
	return p - d;
    }
    else {
	return 0;
    }
}

static int
extract_file_root(d)
    const char *d;
{
    int f = extract_file_drive(d);
    if (isdirsep(d[f])) ++f;
    return f;
}

static void
conv_separator(p)
    char *p;
{
    while (*p) {
	if (*p == '\\') {
	    *p++ = '/';
	}
	else {
	    p = CharNext(p);
	}
    }
}

static VALUE
rb_file_s_drivename(klass, fname)
    VALUE klass, fname;
{
    char *name;
    int f;
    VALUE drivename;

    name = StringValuePtr(fname);
    f = extract_file_drive(name);
    drivename = rb_str_new(name, f);
    conv_separator(RSTRING(drivename)->ptr);

    OBJ_INFECT(drivename, fname);
    return drivename;
}

static VALUE
rb_file_s_rootdir(klass, fname)
    VALUE klass, fname;
{
    char *name;
    int f;
    VALUE rootdir;

    name = StringValuePtr(fname);
    f = extract_file_root(name);
    rootdir = rb_str_new(name, f);
    conv_separator(RSTRING(rootdir)->ptr);

    OBJ_INFECT(rootdir, fname);
    return rootdir;
}

static VALUE
rb_file_s_dirname(klass, fname)
    VALUE klass, fname;
{
    char *name, *p;
    int d;
    VALUE dirname;

    name = StringValuePtr(fname);
    d = extract_file_root(name);
    p = strrdirsep(name + d);
    if (p)
	d = p - name;
    else if (d == 0)
	return rb_str_new2(".");
    dirname = rb_str_new(name, d);
    conv_separator(RSTRING(dirname)->ptr);
    OBJ_INFECT(dirname, fname);
    return dirname;
}

static int
rmext(p, e)
    const char *p, *e;
{
    int l1, l2;

    l1 = strlen(p);
    if (!e) return 0;

    l2 = strlen(e);
    if (l2 == 2 && e[1] == '*') {
	e = strrchr(p, *e);
	if (!e) return 0;
	return e - p;
    }
    if (l1 < l2) return l1;

    if (strcmp(p+l1-l2, e) == 0) {
	return l1-l2;
    }
    return 0;
}

static VALUE
rb_file_s_basename(argc, argv)
    int argc;
    VALUE *argv;
{
    VALUE fname, fext, basename;
    char *name, *p, *ext;
    int d, f;

    if (rb_scan_args(argc, argv, "11", &fname, &fext) == 2) {
	ext = StringValuePtr(fext);
    }
    name = StringValuePtr(fname);
    d = extract_file_root(name);
    name += d;
    p = strrdirsep(name);
    if (!p) {
	if (NIL_P(fext) || !(f = rmext(name, ext))) {
	    if (!d) return fname;
	    basename = rb_str_new2(name);
	}
	else {
	    basename = rb_str_new(name, f);
	}
    }
    else {
	p++;			/* skip last `/' */
	if (NIL_P(fext) || !(f = rmext(p, ext))) {
	    basename = rb_str_new2(p);
	}
	else {
	    basename = rb_str_new(p, f);
	}
    }
    OBJ_INFECT(basename, fname);
    return basename;
}

static VALUE
rb_file_s_split(klass, path)
    VALUE klass, path;
{
    return rb_assoc_new(rb_file_s_dirname(Qnil, path), rb_file_s_basename(1,&path));
}

#ifdef HAVE_SYS_PARAM_H
# include <sys/param.h>
#else
# define MAXPATHLEN 1024
#endif

static VALUE
rb_str_uniqsep_bang(str)
    VALUE str;
{
    long i;
    int d;
    int slash;
    char *p;

    int len;
    char *ptr;

    /*
     *  gsub //+ to / except ^//
     */
    rb_str_modify(str);
    len = RSTRING(str)->len;
    ptr = RSTRING(str)->ptr;
    conv_separator(ptr);
    d = extract_file_drive(ptr);
    slash = 0;
    p = ptr + d;
    for (i=d; i<len; i++) {
        if (slash)
	    if (ptr[i] == '/')
		continue;
	    else
		slash = 0;
	else
	    if (ptr[i] == '/')
		slash = 1;
	*p++ = ptr[i];
    }
    rb_str_resize(str, p - ptr);
    return str;
}

static VALUE
rb_ary_strlist(ary)
    VALUE ary;
{
    long i;
    VALUE ret, tmp;

    ret = rb_ary_new2(RARRAY(ary)->len);
    for (i=0; i<RARRAY(ary)->len; i++) {
	tmp = RARRAY(ary)->ptr[i];
	switch (TYPE(tmp)) {
	  case T_STRING:
	    tmp = rb_str_dup(tmp);
	    rb_ary_push(ret, tmp);
	    break;
	  case T_ARRAY:
	    tmp = rb_ary_strlist(tmp);
	    rb_ary_concat(ret, tmp);
	    break;
	  default:
	    tmp = rb_obj_as_string(tmp);
	    rb_ary_push(ret, tmp);
	}
    }
    return ret;
}

static VALUE separator;

static VALUE
rb_file_s_join(klass, args)
    VALUE klass, args;
{
    long i;
    int d;
    VALUE ary, str;

    ary = rb_funcall2(args, rb_intern("flatten"), 0, 0);
    for (i=0; i<RARRAY(ary)->len; i++) {
	RARRAY(ary)->ptr[i] = str = rb_str_dup(RARRAY(ary)->ptr[i]);
	if (RSTRING(str)->len > 0) {
	    conv_separator(RSTRING(str)->ptr);
	    d = extract_file_root(RSTRING(str)->ptr);
	    if (d == RSTRING(str)->len && RSTRING(str)->ptr[d-1] == '/') {
		/* chop! */
		RSTRING(str)->len--;
	    }
	}
    }
    str = rb_ary_join(ary, separator);
//    str = rb_str_uniqsep_bang(str);
    return str;
}

void
Init_dosish()
{
    rb_mDOSish = rb_define_module("DOSish");
    rb_define_module_function(rb_mDOSish, "drivename", rb_file_s_drivename, 1);
    rb_define_module_function(rb_mDOSish, "rootdir", rb_file_s_rootdir, 1);
    rb_define_module_function(rb_mDOSish, "dirname", rb_file_s_dirname, 1);
    rb_define_module_function(rb_mDOSish, "basename", rb_file_s_basename, -1);

    separator = rb_str_new2("/");
    rb_define_const(rb_mDOSish, "Separator", separator);
    rb_define_const(rb_mDOSish, "SEPARATOR", separator);
    rb_define_singleton_method(rb_mDOSish, "split",  rb_file_s_split, 1);
    rb_define_singleton_method(rb_mDOSish, "join",   rb_file_s_join, -2);

//    rb_define_method(rb_cArray, "strlist",   rb_ary_strlist, 0);
}
